/*
 * Copyright (c) 2004 Serge Vakulenko <vak@cronyx.ru>
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <fcntl.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <sysexits.h>
#include <pwd.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define PATH_SENDMAIL	"/usr/sbin/sendmail"
#define PATH_CLAMD	"/usr/local/sbin/clamd"
#define PATH_CLAMD_PID	"/var/virus/clamd.pid"
#define PATH_KAVD	"/usr/local/share/AVP/kavdaemon"
#define PATH_KAVD_PID	"/var/run/kavdaemon.pid"
#define PATH_AVED	"/usr/local/share/kav/bin/aveserver"
#define PATH_AVED_PID	"/var/run/aveserver.pid"

extern int optind;
extern char *optarg;

void (*outlog) (int, const char *, ...);

/*
 * Write message to standard error
 *
 * Append a message to the standard error for the convenience of
 * end-users debugging without access to the syslog messages.
 */
static void
debuglog (int priority, const char *message, ...)
{
	va_list ap;

	va_start (ap, message);
	vfprintf (stderr, message, ap);
	va_end (ap);
	fprintf (stderr, "\n");
}

static void fatal (const char *message, ...)
{
	va_list ap;

	va_start (ap, message);
	outlog (LOG_ERR, message, va_arg(ap,void*));
	va_end (ap);
	exit (EX_TEMPFAIL);
}

/*
 * Store message in a temporary file.
 */
char *
store_message (char *quarantine_dir)
{
	time_t now;
	const struct tm *tm;
	static char filename [128];
	int fildes, c;
	FILE *fd;

	/* Store different days in different directories
	 * to make them easier to manage. */
	now = time (0);
	tm = localtime (&now);

	sprintf (filename, "%s/%04d-%02d-%02d", quarantine_dir,
		tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday);
	if (mkdir (filename, 0700) < 0 && errno != EEXIST)
		fatal ("mkdir %s failed", filename);

	/* Put time into file name for convenience. */
	sprintf (filename + strlen (filename), "/%02d-%02d-XXXXXX",
		tm->tm_hour, tm->tm_min);
#ifdef HAVE_MKSTEMP
	fildes = mkstemp (filename);
#else
	if (! mktemp (filename))
		fatal ("mktemp %s failed", filename);
	fildes = open (filename, O_CREAT | O_EXCL | O_WRONLY | O_TRUNC, 0600);
#endif
	if (fildes < 0)
		fatal ("cannot create file %s", filename);

	fd = fdopen (fildes, "w");
	if (! fd)
		fatal ("cannot reopen file descriptor");

	/* Put message here. */
	while ((c = getc (stdin)) >= 0)
		putc (c, fd);
	fclose (fd);
	return filename;
}

void
restart_server (char *executable, char *argv0, char *pidfile)
{
	FILE *fd;
	int pid;

	fd = fopen (pidfile, "r");
	if (fd) {
		/* Kill daemon before restart. */
		pid = 0;
		fscanf (fd, "%d", &pid);
		fclose (fd);

		if (pid > 1 && kill (pid, 15) >= 0) {
			sleep (2);
			if (kill (pid, 15) >= 0)
				sleep (2);
		}
	}
	pid = fork ();
	if (pid < 0)
		fatal ("cannot fork");

	if (! pid) {
		/* We are a child. */
		daemon (0, 0);
		execl (executable, argv0, (void*) 0);
		fatal ("restarting %s failed", executable);
	}
}

char *
clamav_check (char *server, char *filename)
{
	FILE *fd;
	static char buf [2048];
	struct iovec iovec [2];
	int sockd, retry;
	int port = 3310;
	char *p, *q;

	/* Connect to clamd server. */
	if (*server == '/') {
		struct sockaddr_un addr;

		strncpy (addr.sun_path, server, sizeof (addr.sun_path));
		addr.sun_family = AF_UNIX;

		sockd = socket (AF_UNIX, SOCK_STREAM, 0);
		if (sockd < 0)
			fatal ("cannot create unix socket");
		for (retry=0; ; ++retry) {
			if (connect (sockd, (struct sockaddr*) &addr,
			    sizeof (addr)) >= 0)
				break;
			if (retry >= 10)
				fatal ("cannot connect to clamd %s", server);
			outlog (LOG_ERR, "restarting clamd %s", PATH_CLAMD);
			restart_server (PATH_CLAMD, PATH_CLAMD, PATH_CLAMD_PID);
			sleep (5);
		}
	} else {
		struct hostent *he;
		struct sockaddr_in addr;

		p = strchr (server, ':');
		if (p) {
			*p++ = 0;
			port = strtol (p, 0, 0);
		}
		he = gethostbyname (server);
		if (! he)
			fatal ("cannot resolve hostname %s", server);
		memcpy (&addr.sin_addr.s_addr, he->h_addr, he->h_length);
		addr.sin_port = htons (port);
		addr.sin_family = AF_INET;

		sockd = socket (PF_INET, SOCK_STREAM, 0);
		if (sockd < 0)
			fatal ("cannot create tcp socket");
		if (connect (sockd, (struct sockaddr*) &addr, sizeof (addr)) < 0) {
			close (sockd);
			fatal ("cannot connect to clamd %s:%d", server, port);
		}
	}

	/* Send SCAN command to clamd. */
	iovec[0].iov_base = "SCAN ";
	iovec[0].iov_len  = 5;
	iovec[1].iov_base = filename;
	iovec[1].iov_len  = strlen (filename);
	if (writev (sockd, iovec, 2) <= 0)
		fatal ("error writing socket");

	/* Read clamd reply. */
	fd = fdopen (sockd, "r");
	if (! fd)
		fatal ("cannot reopen socket descriptor");
	do {
		if (! fgets (buf, sizeof (buf), fd)) {
			/* No errors, no viruses. */
			fclose (fd);
			return 0;
		}
		if (strstr (buf, "ERROR")) {
			/* Server error - strip trailing NL. */
			p = strpbrk (buf, "\r\n");
			if (p)
				*p = 0;
			fclose (fd);
			fatal ("%s", buf);
		}
	} while (! strstr (buf, "FOUND\n"));
	fclose (fd);

	/* Virus found - extract virus name. */
	p = strstr (buf, ": ");
	if (p)
		p += 2;
	else
		p = buf;
	q = strchr (p, ' ');
	if (q)
		*q = 0;
	return p;
}

char *
kav_check (char *server, char *filename)
{
	int sockd, n, retry;
	unsigned char exit_code[2];
	char *reply, *p;
	struct sockaddr_un addr;
	char request [200];
	unsigned long len;

	strncpy (addr.sun_path, server, sizeof (addr.sun_path));
	addr.sun_family = AF_UNIX;
	sockd = socket (AF_UNIX, SOCK_STREAM, 0);
	if (sockd < 0)
		fatal ("cannot create unix socket");
	for (retry=0; ; ++retry) {
		if (connect (sockd, (struct sockaddr*) &addr,
		    sizeof (addr)) >= 0)
			break;
		if (retry >= 10)
			fatal ("cannot connect to kavdaemon %s", server);
		outlog (LOG_ERR, "restarting kavdaemon %s", PATH_KAVD);
		restart_server (PATH_KAVD, PATH_KAVD, PATH_KAVD_PID);
		sleep (20);
	}

	/* My kav server wants only /tmp* and /home* files.
	 * Use dummy symlink. */
	if (access ("/tmpkav", 0) < 0)
		symlink ("/", "/tmpkav");
	sprintf (request, "<0>Jan  1 00:00:00:\xfeI0|K|LL|O-|s|o{/tmpkav%s}\xfe",
		filename);
	if (write (sockd, request, strlen (request)) < 0)
		fatal ("kaspersky write failed on file %s", filename);

	if (read (sockd, (char*) &exit_code, 2) < 2)	/* read first 2 bytes */
		fatal ("kaspersky read(2) failed on file %s", filename);
	if (exit_code[0] == '0' || exit_code[0] == '8') {
		/* No viruses. */
		/*outlog (LOG_NOTICE, "no viruses: %s", reply);*/
		close (sockd);
		return 0;
	}

	/* Virus detected or something bad happened. */
	if (exit_code[1] != 1) {
		close (sockd);
		return "(UNKNOWN)";
	}

	/* read length of an account buffer and then read the buffer */
	if (read (sockd, (char*) &len, sizeof (unsigned long)) < 0)
		fatal ("kaspersky read(4) failed on file %s", filename);

	if (len < 1 || len > 50000) {
		close (sockd);
		return "(UNKNOWN)";
	}

	reply = malloc (len + 1);
	if (! reply)
		fatal ("out of memory");

	p = reply;
	while (len > 0) {
		n = read (sockd, p, len);
		if (n < 0)
			fatal ("kaspersky read() failed");
		if (n == 0)
			break;
		p += n;
		len -= n;
	}
	*p = 0;
	close (sockd);
	/*outlog (LOG_NOTICE, "kavdaemon output: %.100s", reply);*/

	if ((p = strstr (reply, "\tinfected: ")))
		p += 10;
	else if ((p = strstr (reply, "    infected: ")))
		p += 13;
	else if ((p = strstr (reply, "\tsuspicion: ")))
		p += 11;
	else if ((p = strstr (reply, "    suspicion: ")))
		p += 14;
	else
		return reply;

	/* Remove leading and trailing spaces. */
	p += strspn (p, "\t ");
	p [strcspn (p, "\n\r\t ")] = 0;
	return p;
}

char *
ave_check (char *server, char *filename)
{
	int sockd, n, retry;
	char *p;
	struct sockaddr_un addr;
	static char request [4096];

	strncpy (addr.sun_path, server, sizeof (addr.sun_path));
	addr.sun_family = AF_UNIX;
	sockd = socket (AF_UNIX, SOCK_STREAM, 0);
	if (sockd < 0)
		fatal ("cannot create unix socket");
	for (retry=0; ; ++retry) {
		if (connect (sockd, (struct sockaddr*) &addr,
		    sizeof (addr)) >= 0)
			break;
		if (retry >= 10)
			fatal ("cannot connect to aveserver %s", server);
		outlog (LOG_ERR, "restarting aveserver %s", PATH_AVED);
		restart_server (PATH_AVED, PATH_AVED, PATH_AVED_PID);
		sleep (20);
	}

	/* Read AVE banner. */
	request[sizeof(request)-1] = 0;
	if (read (sockd, request, sizeof(request)-1) < 0)
		fatal ("kaspersky read() failed on file %s", filename);
	/*outlog (LOG_NOTICE, "aveserver banner: `%s'\n", request);*/
	switch (request[0]) {
	case '2':	/* successfully connected */
		break;
	case '4':	/* AV engine is busy */
		fatal ("kaspersky is busy: %s", request);
	default:	/* any 5xx code, get out */
		fatal ("kaspersky error: %s", request);
	}

	/* Send scan command. */
	sprintf (request, "SCAN xmPQRSTUWbgh %s\r\n", filename);
	if (write (sockd, request, strlen (request)) < 0)
		fatal ("kaspersky write failed on file %s", filename);

	/* Read reply. */
	n = read (sockd, request, sizeof(request)-1);
	if (n < 0)	/* read first 2 bytes */
		fatal ("kaspersky read() failed on file %s", filename);
	request[n] = 0;
	/*outlog (LOG_NOTICE, "reply: %s\n", request);*/

	close (sockd);

	if (request[0] == '5')
		fatal ("kaspersky error: %s", request);

	if (request[0] != '2' && request[0] != '3')
		return 0;

	request[3] = 0;
	switch (atoi (request)) {
	case 221:
	case 230:
	case 231:
	case 232:
	case 233:
		return "(UNKNOWN)";
	case 322:
	case 327:
	case 328:
		p = strchr (request+4, ' ');
		if (p)
			*p = 0;
		return request+4;
	}
	return 0;
}

/*
 * Run sendmail to send the message.
 */
FILE *
run_mailer (char *to, char *from)
{
	FILE *pipefd;
	int pid;
	int pvect [2];
	char *argv [8];

	if (pipe (pvect) < 0)
		fatal ("pipe: %s", strerror(errno));
	argv[0] = "sendmail";
	argv[1] = "-oi";
	argv[2] = "-f";
	argv[3] = from;
	argv[4] = "--";
	argv[5] = to;
	argv[6] = 0;
	pid = fork();
	if (pid < 0)
		fatal ("fork: %s", strerror(errno));
	if (pid == 0) {
		dup2 (pvect[0], 0);
		close (pvect[0]);
		close (pvect[1]);
		execv (PATH_SENDMAIL, argv);
		fatal ("cannot exec sendmail: %s", strerror(errno));
	}
	/* check return status of the following calls? XXX */
	close (pvect[0]);
	pipefd = fdopen (pvect[1], "w");
	if (! pipefd)
		fatal ("cannot open pipe to sendmail");
	return pipefd;
}

/*
 * Send virus report.
 */
void
send_report (char *reportaddr, char *fromaddr, char *virusname, char *screener,
	char *filename, char *to, char *from, char *host)
{
	FILE *fd, *pipefd;
	char buf [512];

	fd = fopen (filename, "r");
	if (! fd)
		fatal ("no %s file", filename);
	pipefd = run_mailer (reportaddr, fromaddr);
	fprintf (pipefd, "From: %s\n", fromaddr);
	fprintf (pipefd, "To: %s\n", reportaddr);
	fprintf (pipefd, "Subject: %s: Virus from %s\n", screener, host);
	fprintf (pipefd, "Auto-Submitted: auto-generated (antivirus notify)\n\n");

	fprintf (pipefd, "*** VIRUS DETECTED: %s ***\n\n", virusname);
	fprintf (pipefd, "The message has been blocked by %s.\n\n", screener);
	fprintf (pipefd, "To: %s\n", to);
	fprintf (pipefd, "From: %s\n", from);
	fprintf (pipefd, "Host: %s\n", host);
	fprintf (pipefd, "Saved as: %s\n", filename);
	fprintf (pipefd, "--------------------------------------------------------------\n");
	while (fgets (buf, sizeof buf, fd)) {
		if (*buf == '\n' || *buf == '\r')
			break;
		fputs (buf, pipefd);
	}
	fprintf (pipefd, "--------------------------------------------------------------\n");
	fclose (pipefd);
	fclose (fd);
}

/*
 * Send sane message file.
 */
void
send_message (char *to, char *from, char *filename)
{
	FILE *fd, *pipefd;
	int c;

	fd = fopen (filename, "r");
	if (! fd)
		fatal ("no %s file", filename);
	pipefd = run_mailer (to, from);
	while ((c = getc (fd)) >= 0)
		putc (c, pipefd);
	fclose (pipefd);
	fclose (fd);
}

int
main (int argc, char **argv)
{
	int ch;
	char *fromaddr = "<>";
	char *user = "clamav";
	char *reportaddr = "postmaster";
	char *clamserver = 0;
	char *kavserver = 0;
	char *aveserver = 0;
	char *quarantine_dir = "/var/virus";
	char *virusname = 0;
	char *from, *to, *host, *filename, *screener;
	struct passwd *pw;

	openlog("screener", LOG_PID, LOG_MAIL);
	outlog = isatty (2) ? debuglog : syslog;

	while ((ch = getopt (argc, argv, "c:df:k:q:r:u:a:")) != -1) {
		switch (ch) {
		case 'c':
			clamserver = optarg;
			break;
		case 'd':			/* debug mode */
			outlog = debuglog;
			break;
		case 'f':
			fromaddr = optarg;
			break;
		case 'k':
			kavserver = optarg;
			break;
		case 'a':
			aveserver = optarg;
			break;
		case 'q':
			quarantine_dir = optarg;
			break;
		case 'r':
			reportaddr = optarg;
			break;
		case 'u':
			user = optarg;
			break;
		default:
usage:			outlog (LOG_NOTICE, "usage: screener "
				"[-d] [-r report-addr] [-f from-addr] [-c clamserver] [-a aveserver] [-k kavdaemon] [-u user] [-q dir] "
				"to from host");
			exit (EX_TEMPFAIL);
		}
	}
	argc -= optind;
	argv += optind;

	if (argc != 3)
		goto usage;
	to = argv[0];
	from = argv[1];
	host = argv[2];

	pw = getpwnam (user);
	if (! pw)
		fatal ("getpwnam(): %s", strerror(errno));

	/* Store message in file, owned by user. */
	if (seteuid (pw->pw_uid) < 0)
		fatal ("seteuid(0): %s", strerror(errno));
	filename = store_message (quarantine_dir);
	if (seteuid (0) < 0)
		fatal ("seteuid(0): %s", strerror(errno));

	if (clamserver) {
		virusname = clamav_check (clamserver, filename);
		if (virusname) {
			screener = "clamav";
			goto infected;
		}
	}
	if (aveserver) {
		virusname = ave_check (aveserver, filename);
		if (virusname) {
			screener = "kaspersky";
			goto infected;
		}
	}
	if (kavserver) {
		virusname = kav_check (kavserver, filename);
		if (virusname) {
			screener = "kaspersky";
			goto infected;
		}
	}

	/* Message is sane. */
	outlog (LOG_INFO, "approve message to %s from %s host %s", to, from, host);
	send_message (to, from, filename);
	unlink (filename);
	return EX_OK;
infected:
	outlog (LOG_NOTICE, "virus %s from %s thanks %s",
		virusname, host, screener);
	send_report (reportaddr, fromaddr, virusname, screener,
		filename, to, from, host);
	return EX_OK;
}
